'''
 This file is generated form: gen/__init__.py.mako on 2024-07-11 15:05:08.891320 by calling E:/1_ILS/Projects/2018_EOQ/Groups/PYEOQ/pyeoq/eoq3conceptsgen/generatefromconceptscli.py -i gen/__init__.py.mako -o __init__.py.
 
 Bjoern Annighoefer 2023
'''

from .concepts import\
    CONCEPTS,\
    MXMDB,\
    MXELEMENT,\
    MXCONSTRAINT,\
    M2PRIMITIVES,\
    M2PACKAGE,\
    M2MODEL,\
    M2ENUM,\
    M2OPTIONOFENUM,\
    M2CLASS,\
    M2ATTRIBUTE,\
    M2ASSOCIATION,\
    M2COMPOSITION,\
    M2INHERITANCE,\
    M1MODEL,\
    M1OBJECT,\
    M1FEATURE,\
    M1ATTRIBUTE,\
    M1ASSOCIATION,\
    M1COMPOSITION,\
    CONCEPT_PREFIX,\
    CONCEPT_PREFIX_LEN,\
    CONCEPT_UNIQUE_LEN,\
    FEATURE_MULTIVALUE_POSTFIX,\
    IsConcept,\
    IsMultivalueFeature,\
    NormalizeFeatureName,\
    GetConceptKeyString
